#!/usr/bin/python3

import threading, queue, time, traceback, sys, subprocess
from . import signals


"""
Launch functions in threads with collated logging. Protection is offered for non-thread
safe operations.

For the tl;dr crowd, here's a quickstart. All you need are three lines of code, one
line per thread you want to launch, plus a function definition to run:

    from coolthreads import CoolThreads

    # define a function you want to run one or more copies of. This one prints
    two messages, one safely and one not, and makes a log entry:

    def testAdHoc(**kwargs):
        print(f'not thread-safe from ad hoc thread {kwargs["comment"]}')
        CT = kwargs['CT'] # CT is always there, points back to CoolThread instance
        CT.CTthreadSafe(CT.getName(), print, f'thread from ad hoc thread {kwargs["comment"]}')
        CT.CTlogger(f'log entry from ad-hoc {CT.getName()}', CT.getName())

    # create a dispatcher

    qControl = CoolThreads('myThreadName')
    qControl.start()

    # run a few copies

    qControl.CTaddThread('adhoc-1', CoolThreadOp('adhoc-1', CTrunFunction = testAdHoc, comment = f'thread adhoc-1 is king of the world'))
    qControl.CTaddThread('adhoc-2', CoolThreadOp('adhoc-2', CTrunFunction = testAdHoc, comment = f'no, thread adhoc-2 is king of the world'))
    qControl.CTaddThread('adhoc-3', CoolThreadOp('adhoc-3', CTrunFunction = testAdHoc, comment = f'no, thread adhoc-3 is king of the world'))
    qControl.CTaddThread('adhoc-4', CoolThreadOp('adhoc-4', CTrunFunction = testAdHoc, comment = f'no, thread adhoc-4 is king of the world'))

    # wait for the dust to settle and terminate

    qControl.CTwaitForComplete()

CoolThreads is a dispatcher, running in a thread. It manages threads running as class CoolThreadOp.

Python's main thread remains available for use.

Constructor arguments to CoolThreads are
    thrdName - name for Thread class name.
    logFile - defaults to None, a file object for logging. If None, logs to console.
    threadMax - maxiumum number of threads to run at any one time.
    collate - defaults to True. This causes log output to be collated by thread.
    queueSleep - defaults to zero seconds. Time to sleep on finding an empty log/command queue.
        Can be zero, but may provide some performance improvement by sleeping the main
        thread when there is likely little to do. If the individual threads are expected to run
        quickly, this value should be zero.
    consoleLog - boolean option to duplicate log output to console. Defaults to True.
    kwargs - additional keyword arguments that are passed to CTrun (see below) and to 
        CoolThreads internal functions such as CTcleanThreads. CTcleanThreads makes no use of
        kwargs, but user supplied overrides can.

Methods and attributes in CoolThreads are named in the form CTsomethingInteresting. This is to
avoid name collision with inheritance and user code.

CoolThreads inherits two classes, threading.Thread and signals.SigIntercept. Signals.SigIntercept
is part of this package.

Use CoolThreads by inheriting it into a class. This gives the implementer full access to all 
resources in CoolThreads and its superclasses.

A single new member is required for CoolThreads to run, a function called CTrun. This function is 
called in each new thread. 

Once an instance of CoolThreads is created, call it's start function. Start is inherited from 
threading.Thread. No processing will really begin, that just starts the engine.

CTaddThread allows parameterized thread instances to be initiated. It takes two arguments, the
name of the thread being added and a reference to the thread object to add which should be a
subclass of CoolThreadOp. Requiring the thread name as a parameter may seem redundant, since 
the name is part of CoolThreadOp. However, in the case of an exception during adding the thread,
nothing in the new object is trusted. For logging purposes, an independent copy of the name
is needed.

if CoolThreadOp is passed an argument of CT = None (or any value), it will suppress adding an
argument setting kwargs['CT']. This allows use of functions that won't tolerate unexpected command
arguments.

CTlogger takes three arguments, a line (or lines) to be logged, an optional threadid string
and a bit mapped command. See comments below on bit map command/logging options.

CTcleanThreads is the cleanup method called by SigIntercept (executed on control-C). The included
implementation of this method shuts off new threads by setting CTopenFoBiz to False.

CTlogger takes a log line (or lines), a threadId, and a command bitfield.

CTdumpThreadLog logs the CTjobLog dictionary for a specific thread.

CTdumpLog logs all entries in CTjobLog.

CTjobLog is a dictionary. The key is thread id, data is a list of log lines. This is for deferred
logging when the collate option is set in the constructor for CoolThreads.

CoolThreads.run is used by threading.Thread. It watches for command and log entries
appearing in CTlogQueue.

CTnoThread is a resource lock to allow something to run without interference from other threads. This
uses a threading.RLock object, so the locking is at thread level, not at resource level. No blocking
conditions occur if a thread tries to acquire the lock from different points of execution within a
thread.

CTthreadSafe takes two required arguments, threadID and targFunction. Other arguments are passed to 
targFunction. If targFunction raises an exception, its thread will be terminated. Other processing
will continue.

CTprintConsole takes two required arguments, a threadid for logging and a string to print on the console in thread-safe conditions.
A second argument is True if user input is required. A third argument, if present, is a prompt if the
user's input doesn't meet expectations, and a fourth optional argument is a list of allowed input strings.
Note that a string like 'yn' appears to be useful, because either 'y' or 'n' by itself will pass the test
of being in the allowed input list. 'yn' will also pass, as will ''.

User responses are logged.

CTstop - turns off thread processing. The CoolThreads run loop will terminate when all threads have completed.

CTterminated is a boolean that is set when all logging and processing is finished.

CTopenForBiz is a boolean that is set if CoolThreads is available for adding threads.

A simple example is in the if __name__ == '__main__' section of this file. If you run this within the 
package directory, you may need to temporarily change:
    from . import signals
to:
    import signals
"""

# bit map command/logging options
# least byte, job and threading options
# next most significant byte, log options
# b0 - continue threading
# b1 - flag job stop
# b2 - threading stop request
# b8 - console message only (always immediate)
# b9 - write log immediately
# b10 - write deferred log
CTcontinueThreading = 1
CTjobStop = 2
CTthreadingStop = 4
CTconsoleMsg = 0x100
CTimmediateLog = 0x200
CTdeferredLog = 0x400
CTimmediateDeferred = 0x600

class CoolThreads(threading.Thread, signals.SigIntercept):

    def __init__(self, thrdName, logFile = None, threadMax = 10, collate = True, queueSleep = 0, consoleLog = True, **kwargs):
        threading.Thread.__init__(self, name = thrdName) # initialize threading
        signals.SigIntercept.__init__(self, cleanup = self.CTcleanThreads) # catch control-c
        self.CTlogFile = logFile # CTlogQueue is dumped to this file, if not None
        self.CTthreadMax = threadMax # the maxiumum number of threads to run
        self.CTcollate = collate # option flag (boolean) to collate output by thread
        self.CTqueueSleep = queueSleep # optional time to sleep on finding an empty command/log queue
        self.CTconsoleLog = consoleLog
        self.kwargs = kwargs # args to pass to internal and user-supplied functions
        self.CTthreadLimit = threading.BoundedSemaphore(threadMax) # semaphone in the main thread to limit thread count
        self.CTnoThread = threading.RLock() # mutex to allow thread protection for things like writing to stdout
        self.CTlogQueue = queue.Queue() # command/log queue for threads to report status and logging
        self.CTjobLog = {} # key is threadId, data is list of log lines
        self.CTterminated = False
        self.CTopenForBiz = True
        if collate: # set logging options
            self.CTstandardLogMethod = CTdeferredLog
        else:
            self.CTstandardLogMethod = CTimmediateLog
        if consoleLog:
            self.CTstandardLogMethod |= CTconsoleMsg
        self.CTstandardLogMethod |= CTcontinueThreading # normal logging always includes the option for thread dispatch to continue
        self.start()

    def CTcleanThreads(self, sig, frame, **kwargs):
        self.CTopenForBiz = False # executed from SigIntercept for control c
        print('*** Control-C caught *** New threads will be rejected')

    def CTlogger(self, logLine, threadId = None, cmd = None):
        if cmd is None:
            cmd = self.CTstandardLogMethod # default to standard logging
        logLine = logLine.strip() # newlines are stripped and added back for file logging
        immediateLogLine = logLine # this is the version of the logLine written to the console
        if threadId:
            immediateLogLine = f'{time.strftime("%Y-%m-%d %H:%M:%S")} {threadId}: {logLine}'
        if cmd & CTconsoleMsg: # cmd is a bit map
            self.CTprintConsole(threadId, immediateLogLine + '\n')
        if (cmd & CTimmediateLog) and self.CTlogFile: # write to log file if immediate logging is requested (not collated) and logFile exists
            self.CTlogFile.write(immediateLogLine + '\n')
            return
        if cmd & CTdeferredLog: # save log lines in a dictionary for collated output
            if not threadId:
                threadId = 'Main thread'
            if not threadId in self.CTjobLog:
                self.CTjobLog[threadId] = []
            self.CTjobLog[threadId].append(logLine)

    def CTdumpThreadLog(self, threadId):
        if threadId in self.CTjobLog: # dump the log for this thread, if it exists
            for l in self.CTjobLog[threadId]:
                self.CTlogger(l, threadId, CTimmediateLog) # immediate logging - requires logFile to exist

    def CTdumpLog(self): # dump all thread logging
        for threadId in self.CTjobLog:
            self.CTdumpThreadLog(threadId)

    def CTactiveCount(self):
        return(self.CTthreadMax - self.CTthreadLimit._value)

    def run(self):
        while self.CTopenForBiz or self.CTactiveCount(): # while either accepting new threads or threads are running
            # serve logging and command queue
            try:
                logQ = self.CTlogQueue.get(block = False) # check for log/command data - raises exception if nothing is there
                cmd, threadId, logLine = logQ # queue is bit mapped command and logging options, thread id, and text to log
                self.CTlogger(logLine, threadId, cmd) # log the data as approprate in setup
                if cmd & CTthreadingStop:
                    self.CTopenForBiz = False # on stop request, refuse additional thread additions
                if cmd & CTjobStop: # if job is stopping, dump its log
                    self.CTdumpThreadLog(threadId)
                    if threadId in self.CTjobLog:
                        del(self.CTjobLog[threadId])
                    try:
                        self.CTthreadLimit.release() # release the queue limit lock, trap excessive release errors
                    except ValueError:
                        self.CTopenForBiz = False
                        self.CTlogQueue.put((CTconsoleMsg | CTthreadingStop, threadId, 'Semaphore release exceeded boundary - terminating'))
            except queue.Empty: # catch exceptions from queue get operation fail
                if self.CTqueueSleep:
                    time.sleep(self.CTqueueSleep) # optionally wait before checking again
        self.CTdumpLog() # after shutdown request and no more threads are running, dump deferred log
        self.CTterminated = True # flag threading is complete including logging

    def CTthreadSafe(self, threadId, targFunction, *args, **kwargs): # run something with thread privacy
        with self.CTnoThread: # acquire lock for one-at-a-time execution
            try:
                return targFunction(*args, **kwargs) # attempt passed function
            except Exception as e: # log problem
                self.CTlogger(f'Exception in threadsafe call:\n{str(e)}', threadId, self.CTstandardLogMethod | CTconsoleMsg)
                raise # pass exception to caller, which will eventually release this thread's limit semaphore

    def CTprintConsole(self, threadId, outstr, getInput = False, rePrompt = None, allowedInput = None):
        with self.CTnoThread: # acquire lock for console interaction
            sys.stdout.write(outstr) ; sys.stdout.flush()
            if getInput:
                while True:
                    userIn = sys.stdin.readline().strip().lower()
                    if allowedInput and (userIn in allowedInput): break
                    if rePrompt:
                        sys.stdout.write(rePrompt)
                    else:
                        sys.stdout.write(outstr)
                    sys.stdout.flush()
                self.CTlogger(f'User response, {outstr}: {userIn}', threadId, self.CTstandardLogMethod & CTimmediateDeferred)
                return userIn

    def CTaddThread(self, threadId, newThread):
        if not self.CTopenForBiz: return False
        self.CTthreadLimit.acquire() # wait until a sempaphore is available
        try:
            newThread.CT = self # add the CoolThreads reference to the new thread object
            newThread.start() # start the new thread's processing
            return True
        except Exception as e:
            self.CTthreadLimit.release() # the acquire was before the try block, so release the unneeded semaphore
            self.CTlogger(f'Exception in CTaddThread call:\n{str(e)}', threadId, self.CTstandardLogMethod | CTconsoleMsg)
            return False

    def CTannounceTerminate(self, threadId): # queue CTjobStop to get the job's semaphore released
        self.CTlogQueue.put((CTjobStop, threadId, ''))

    def CTstop(self): # request threading shutdown
        self.CTlogQueue.put((CTthreadingStop, None, ''))

    def CTwaitForComplete(self): # request threading shutdown, if that hasn't already been done, wait for everything to complete - must be called from main thread!
        self.CTstop()
        while not self.CTterminated:
            time.sleep(1)

"""
CoolThreadOp is the class that runs the process of interest for the user.
CoolThreads is a dispatcher. This is the target code.

A name is required for the constructor. Optional **kwargs are
saved for the user's thread for any purpose.

By default (without overriding CTrun), CoolThreadOp will run the command passed in constructor kwargs as CTrunFunction.

Note that if CT is already in kwargs, it will be dropped.

If CT is not in kwargs, it will be added as a reference to the controlling CoolThreads dispatcher class.

This supports CTrunFunction procedures that will not tolerate extra arguments.

So, if you're calling a function that would know what to do with kwargs['CT'], let CoolThreadOp.CTrun set that for you.

If you're calling a function that can't tolerate extra arguments, it wouldn't know what to do with kwargs['CT'], so pass an
argument CT = None to CoolThreadOp's constructor.

If argument CTlock is set True, then the thread lock will be acquired before this thread runs.

Note that 'CT' arguments to CoolThreadOp's constructor must be keyword-style, as in "CTlock = True", and not just "True".
"""
class CoolThreadOp(threading.Thread):

    def __init__(self, opName, *args, CTlock = False, **kwargs):
        super().__init__()
        self.opName = opName
        self.args = args
        self.CTlock = CTlock
        self.CTkwargs = kwargs

    def run(self): # run proccess for threading.Thread. self.CTrun must be provided through inheritance
        try:
            if self.CTlock:
                self.CT.CTnoThread.acquire() # will release automatically on thread termination
            self.CTrun(*self.args, **self.CTkwargs)
            if self.CTlock:
                self.CT.CTnoThread.release()
            self.CT.CTannounceTerminate(self.getName()) # announce termination
        except Exception as e:
            self.CT.CTlogger(f'Exception: {str(e)}', self.opName, self.CT.CTstandardLogMethod) # log any exception
            self.CT.CTlogger(traceback.format_exc(), self.opName, self.CT.CTstandardLogMethod)
            self.CT.CTannounceTerminate(self.getName()) # announce thread termination, because untrapped exceptions terminate threads uncleanly

    def CTrun(self, *args, **kwargs):
        if 'CT' in kwargs:
            kwargs.pop('CT')
        else:
            kwargs['CT'] = self.CT
        ctRunF = kwargs['CTrunFunction']
        kwargs.pop('CTrunFunction')
        self.CT.CTlogger(f'Return code: {ctRunF(*args, **kwargs)}', self.opName, self.CT.CTstandardLogMethod)

    def start(self): # sets thread name, allows for startInit hook, and starts this object's threading
        self.setName(name = f'{self.CT.getName()}-{self.opName}')
        self.startInit()
        threading.Thread.start(self)

    def startInit(self):
        pass

def CTthreadExternal(*args, **kwargs):
    cmdParams = []
    for a in args:
        cmdParams.append(a)
    for a in kwargs:
        cmdParams.append(f'{a}={kwargs[a]}')
    p = subprocess.run(cmdParams, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    sOut = p.stdout.decode().strip()
    sErr = p.stderr.decode().strip()
    return(f'stdout: {sOut}\nstderr: {sErr}')

if __name__ == '__main__':
    import random

    """
    A quick example. First set up a class for the threads.

    This class inherts coolthreads.CoolThreadOp. That's most of the work of threading.

    A method called CTrun has to be written. It's just the code to run whatever the threads
    are going to do. It can, of course, call other functions, so the thread instances don't
    have to all do the same thing.

    They all start from CTrun, but other than that there is no dependence between threads.
    """
    class TestOp(CoolThreadOp):

        def printHello(self, **kwargs): # this is a sample independent function, intended to be called by CTthreadSafe so stdout doesn't get mixed up
            print(f'Thread: {self.getName()}, comment: {kwargs["comment"]}, delay: {self.TOdelay}')
            sys.stdout.flush() # all of this function is run thread-protected

        def CTrun(self, **kwargs): # the work of the thread. thread operations can be somewhat ignored at this level
            self.CT.CTlogger('startup', self.getName()) # a call to CTlogger, announcing startup with thread id
            self.TOdelay = int((random.random() * 2) + 3) # set up a random delay. it's a class member so printHello can access it
            self.CT.CTthreadSafe(self.getName(), self.printHello, comment = 'hello there') # call printHello with a comment argument
            self.CT.CTlogger(f'hello, log, mid-run, threads running {self.CT.CTactiveCount()}', self.getName()) # another log
            #self.CT.CTlogger('hello, log, division = {4/(self.TOdelay-4)}', self.getName()) # this will demonstrate some exceptions for divide by zero
            time.sleep(self.TOdelay) # wait a random time, unless the previous line divided by zero
            self.CT.CTlogger(f'hello, log, after nap - threads {self.CT.CTactiveCount()}', self.getName()) # another log entry
            self.CT.CTthreadSafe(self.getName(), sys.stdout.write, f'Terminating {self.getName()}\n') # another direct write to the console

    with open('cooltest.log', 'wt') as coollog: # use the above, with a log file
        testQueue = CoolThreads('testqueue', coollog, queueSleep = 0, collate = True, consoleLog = True, threadMax = 200) # create thread dispatcher - a CoolThreads object
        for i in range(10): # launch copies, by default limited to 10 at a time
            testQueue.CTaddThread(f'thread-{i}', TestOp(f'thread-{i}', comment = str(i))) # add a thread with a thread id
        print('Still running in the main thread') # show the main thread is still running while threads are processing
        testQueue.CTwaitForComplete() # call for shutdown and wait until everything stops
    print('We\'re done') # announce program termination
    # note that the log file needed to stay open until after CTwaitForComplete returned

    def testAdHoc(**kwargs):
        CT = kwargs['CT']
        CT.CTthreadSafe(CT.getName(), print, f'from ad hoc thread {kwargs["comment"]}')
        CT.CTprintConsole('ad hoc thread', f'***** CTprintConsole from {kwargs["comment"]} ', True, 'Try again: ', ('y', 'n'))
        CT.CTlogger(f'hello from ad-hoc {CT.getName()}', CT.getName())

    with open('cooltest1.log', 'wt') as coollog:
        testQueue = CoolThreads('testqueue2', coollog)
        for i in range(5):
            testQueue.CTaddThread(f'thread2-{i}', CoolThreadOp(f'thread2-{i}', CTrunFunction = testAdHoc, comment = f'ad hoc run {i}'))
        testQueue.CTaddThread('Doomed', 'Not a thread object')
        for i in range(5):
            testQueue.CTaddThread(f'thread2-{i}', CoolThreadOp(f'thread2-{i}', CTrunFunction = testAdHoc, comment = f'ad hoc run {i}'))
        testQueue.CTwaitForComplete()
    print('test run 2 complete')
